#!/usr/bin/nawk -f 

#/****************************************************************************/
#/*                                                                          */
#/*                     (c) 1997,89,99 by vhs-soft                           */
#/*                                                                          */
#/* $Id: pdweave.nawk,v 1.8 2002/11/15 17:07:28 volker Exp $
#/*                                                                          */
#/* This file is distributed under the GNU Public License                    */
#/*                                                                          */
#/* Volker Simonis  email : simonis@informatik.uni-tuebingen.de              */
#/*                                                                          */
#/****************************************************************************/

function error(errNo) {
  cvs_version = "$Revision: 1.8 $";
  cvs_version = substr(cvs_version, index(cvs_version, " ") + 1);
  cvs_version = substr(cvs_version, 1, index(cvs_version, " "));
  if (errNo == 1) {
    printf("\npdweave error : one argument required.\n\n") }
  else if (errNo == 2) {
    printf("\npdweave error : can't open \"%s\".\n\n", ARGV[1]) }
  else if (errNo == 3) {
    printf("\npdweave error : input file wasn't a \".pd\" file.\n\n") }
  else if (errNo == 4) {
    printf("\npdweave error : can't open source file \"%s\".\n\n",
	   realSourceFile)}
  else if (errNo == 5) {
    printf("\n\\sourceinput error : ");
    printf("You must supply at least 2 parameters.\n\n")}
  else if (errNo == 6) {
    printf("\npdweave error : can't find label \"%s\" in file \"%s\".\n\n", 
            example, realSourceFile)}
  printf("pdweave %s\n\n", cvs_version);
  printf("usage :   pdweave file.pd\n");
  printf("          where file.pd is a latex file with the special\n");
  printf("          \\sourceinput command, which will be replaced by the\n");
  printf("          actual sourcecode an written to a file named file.tex\n");
  exit -1;
}

# This function tries to make a primitive name mangling in order to ensure
# unique labels in the created source listing. Every listing gets a label
# which is composed out of the string 'ProgDoc' the source file name and 
# the tagname. Additionally, characters in this labelname which can cause
# problems in Tex are relpaced by a 'X'.
function makeLabel(tagname) {
  tagname = "ProgDoc" sourceFile tagname;
  gsub(/\\/, "X", tagname);
  gsub(/\//, "X", tagname);
  gsub(/\./, "X", tagname);
  gsub(/_/, "X", tagname);
  gsub(/~/, "X", tagname);
  return tagname;
}

# This function parses the optional arguments common to the
# "\sourcebegin" and "\sourceinput" commands.
function parseArguments() {
  label = 0;
  fontname = 0;
  fontenc = 0;
  fontsize = 0;
  linesep = 0;
  center = 0;
  underline = 0;
  listing = 0;
  listingType = 0;
  linenr = 0;
  type = 0;
  count = 0;
  caption = 0;
  highlighter = 0;
  useLongtable = 0;
  comment=0;
  wrap = 0;
  TAB = 8;
  # All arguments enclosed between apostrophs ('') will be handeled like a
  # single parameter. Inbetween the apostrophs an apostroph character can be
  # used by quoting it with a slash (\').
  # In order to handle this we first replace all strings between apostrophs
  # by placeholders of the form "_____n_____", where 'n' is a number between
  # 0 and 9, and remember the original strings in the array 'repl'.
  while (match($2, /'([^']|\\')*'/)) {
    repl[count] = gensub(/\\'/, "'", "g", substr($2, RSTART+1, RLENGTH-2));#"
    sub(/'([^']|\\')*'/, "_____" count "_____", $2);
    count++;
  }
  nrArgs = split($2, optArg, ",");
  for (i=1;i<=nrArgs;i++) {
    optArg[i] = " " optArg[i];
    nrOfFields = split(optArg[i], optArgVal, "[ \t=]+");
    # If the argument was a string enclosed by apostrophs, we redo the
    # substitution which was done before and restore the old value of the 
    # parameter string.
    if (nrOfFields == 3 && match(optArgVal[3], /_____[0-9]_____/)) {
      optArgVal[3] = repl[gensub(/_____/, "", "g", optArgVal[3])];
    }
    if (index(optArg[i], "label")) label = optArgVal[3];
    if (index(optArg[i], "fontname")) fontname = optArgVal[3];
    if (index(optArg[i], "fontenc")) fontenc = optArgVal[3];
    if (index(optArg[i], "fontsize")) fontsize = optArgVal[3];
    if (index(optArg[i], "linesep")) linesep = optArgVal[3];
    if (index(optArg[i], "listing")) {
      listing = 1;
      if (nrOfFields == 3 && optArgVal[3] == "noUnderline") 
	listingType =  "noUnderline";
    }
    if (index(optArg[i], "linenr")) linenr = 1;
    if (index(optArg[i], "center")) center = 1;
    if (index(optArg[i], "useLongtable")) useLongtable = 1;
    if (index(optArg[i], "underline")) underline = 1;
    if (index(optArg[i], "type")) type = optArgVal[3];
    if (index(optArg[i], "caption")) caption = optArgVal[3];
    if (index(optArg[i], "wrap")) wrap = optArgVal[3];
    if (index(optArg[i], "highlighter")) highlighter = optArgVal[3];
    if (index(optArg[i], "comment")) comment = optArgVal[3];
    if (index(optArg[i], "tab")) TAB = optArgVal[3];
  }
  if (highlighter != 0) {
    CPP2L_CMD = highlighter; 
  }
  else {
    CPP2L_CMD = CPP2L_CMD_DEFAULT;
  }
  if (CPP2L_CMD == "pdhighlight") {
    wrapChar = "\\\037"; 
  }
  else if (index(CPP2L_CMD, "pdlsthighlight")) {
    wrapChar = "&@!&$\\hookleftarrow$@";
  }
}

# This function converts the type argument of the
# "\sourcebegin" and "\sourceinput" commands into a valid 'pdhighlight'
# argument. If the 'type' argument is not supported, 'pdhighlight' will be
# called with the 'text-mode' option.
function typeName(typeArg) {
  if (typeArg == "c") typeArg = "c-mode";
  else if (typeArg == "cpp") typeArg = "cpp-mode";
  else if (typeArg == "java") typeArg = "java-mode";
  else if (typeArg == "text") typeArg = "text-mode";
  else if (typeArg == "xml") typeArg = "xml-mode";
  else if (typeArg == "scm") typeArg = "scheme-mode";
  else if (typeArg == "el") typeArg = "elisp-mode";
  else typeArg = "text-mode";
  return typeArg;
}

# This function returns true if its argument is the token which
# begins a comment in the actual language-mode (determined by the 
# 'type' variable). If 'type' was not set by the user C/C++ mode is assumed.
function beginComment(token) {
  if (comment && token == comment) return 1;
  if (!type) mode = "cpp"; else mode = type;
  if (mode == "cpp" || mode == "C++" || mode == "c" || mode == "C" || 
      mode == "java" || mode == "Java" )
    if (token == "/*" || token == "//") return 1; else return 0;
  else if (mode == "xml")
    if (token == "<!--") return 1; else return 0;
  else if (mode == "scm" || mode == "el" || mode == "Lisp")
    if (token == ";" || token == ";;" || token == ";;;" || token == ";;;;") 
      return 1; else return 0;
  else if (mode == "vb")
    if (token == "'") return 1; else return 0;
  else if (mode == "py" || mode == "Python")
    if (token == "#") return 1; else return 0;
  else if (mode == "text")
    if (token == "#" || token == "//" || token == "-") return 1; else return 0;
  return 0;
}

# This function checks wether 'needle' is the first word which appears in 
# 'line'. 'needle' may thereby be preceeded by an arbitrary amount of 
# whitespace characters (' ' and '\t').
function lineBeginsWith(line, needle) {
  pos = index(line, needle);
  if (pos == 0) return 0;
  if (match(line, /( |\t)*/) == 1 && RLENGTH == pos-1) return 1;
  else return 0;
}

function wrapLine(line) {
  untabbedLine = line;
  # This is an attempt to properly wrap lines which begin with TAB characters.
  # TABs are replaced by 8 (this is hard-coded right now) ESC characters to
  # get the proper length of the line. Later on 8 ESC characters are 
  # substituted back into one TAB character.
  tabString = "";
  for (i = 0; i < TAB; i++) tabString = tabString "\x1B";
  gsub(/\t/, tabString, untabbedLine);
  if (length(untabbedLine) <= wrap) return line;
  if (match(line, /^[[:space:]]*\/\*\?/) ||
      match(line, /^[[:space:]]*&@!&/)) return line;
  tmp1 = substr(untabbedLine, 1, wrap);
  gsub(tabString, "\t", tmp1);
  break1 = match(tmp1, /[[:space:]][^[:space:]]*$/);
  if (break1 < wrap/2) {
    break2 = match(tmp1, /[^[:alnum:]][[:alnum:]]*$/);
    if (break2 > break1) break1 = break2;
  }
  ret1 = substr(line, 1, break1) wrapChar "\n";
  for (i = 1; i < wrap - (length(line) - break1); i++) ret1 = ret1 " ";
  ret1 = ret1 substr(line, break1 + 1);
  return ret1;
}

BEGIN {
  TAB = 8;
  DELTA = 5;
  CPP2L_CMD = "pdhighlight";
  CPP2L_CMD_DEFAULT = "pdhighlight";
  wrapChar = "\\";

  if (ARGC == 1) error(-1);
  if (ARGC == 2 && (ARGV[1] == "-h")) error(-1);
  if (ARGC != 2) error(1);
  tmpFile = "_pdweave.tmp";

#  This doesn't work on Windows* systems, since there is no test command. 
#
#  testFile ="test -r \"" ARGV[1] "\"";
#  if (system(testFile)) error(2);
  outFile = ARGV[1];
  if (sub(/\.vhs$/, ".tex", outFile) != 1)
    if (sub(/\.pd$/, ".tex", outFile) != 1) error(3);
  
  FS = "(\\[|\\]|{|})+"; # We have to quote '\' in the FS string to use it to
  while ((getline < ARGV[1]) == 1) {     # to quote the '[' and ']' character
    countA = 0;
    if (lineBeginsWith($0, "\\sourceinputbase")) {
      if (gsub(/{/, "{") == 1 && gsub(/}/, "}") == 1) {
	basePath=$2;
      }
    }
    else if (lineBeginsWith($0, "\\sourceinput")) {
      sourceinput = $0;
      # This is only a simple heuristic. We read at most 5 lines in order to
      # find all the arguments required for the \sourceinput command.
      for (i = 0; i < DELTA; ++i) {
        # Protect all braces inside strings enclosed ba apostrophs. For details
        # see the 'parseArguments()' function.
        while (match(sourceinput, /'([^']|\\')*'/)) {
          replA[countA] = gensub(/\\'/, "'", "g", 
                                 substr(sourceinput, RSTART, RLENGTH));#"
           sub(/'([^']|\\')*'/, "_____" countA "_____", sourceinput);
          countA++;
        }

	if ((gsub(/{/, "{", sourceinput) != 2) || 
	    (gsub(/}/, "}", sourceinput) != 2)) {
	  if ((getline < ARGV[1]) != 1) error(5);
	  gsub(/^[ \t]*/,""); # Cut trailing whitespace from input line.
	  sourceinput = sourceinput $0;
	}  
      }

      gsub(/^.*$/,sourceinput); # Replace $0 by 'sourceinput'.
      # At this point, $0 may still contain replaced parts!
      print $0;
      if (gsub(/\[/, "[") != gsub(/\]/, "]")) error(5);
      if ((gsub(/{/, "{") != 2) || (gsub(/}/, "}") != 2)) error(5);
      if (gsub(/\[/, "[") == 1) {
        sourceFile = $3;
        example = $4;
        # Here we restore the parts of $2 which may have been replaced before.
        # This is necessary because 'parseArguments()' uses the global variable
        # $2 directly.
        for (i = 0; i < countA; i++) {
          ind = index($2, "_____" i "_____");
          if (ind) $2 = substr($2, 0, ind-1) replA[i] substr($2, ind + 11);
        }
	parseArguments();
      } else if (gsub(/\[/, "[") == 0) {
        sourceFile = $2;
        example = $3;
      } else error(5);


      realSourceFile = basePath sourceFile;
#  This doesn't work on Windows* systems, since there is no test command. 
#
#      testFile ="test -r \"" realSourceFile "\"";
#      if (system(testFile)) error(4);
      oldFS = FS;
      FS = " "; # See docu: "In  the special case that FS is a single space..
      begincount = 0;# Counter of open "// BEGIN" command lines.
      linecount = 0; # Global file line counter.
      lines = 1;     # Line counter for the lines passed to 'pdhighlight'.
      noblank = 0;   # Number of blanks (indentation) of last line.
      blankcount = 10000; # Something just bigger then the length of a line.
      # 'blankcount' counts the maximal number of blanks commen to all the
      # source lines processed. These blanks will finally be stripped of the 
      # lines, in order for the whole code sequenze to appear left justified.
      maxLineWidth = 0; # Counts the number of characters in the longest line. 
      while ((getline < realSourceFile) == 1) {
        linecount++;
	if (example=="ALL") {
            if (NF != 0) { # We compute 'noblank' only for non-empty lines!
              noblank = match($0, /[^ \t]/);
              if (noblank < blankcount) blankcount = noblank;
            }
	    # if (printLine) printf("%s\n", $0) > tmpFile;
            sourcelines[lines++] = $0;          
        }
	else if (beginComment($1) && $2=="BEGIN" && $3==example) {
          linebegin = linecount + 1;
	  printLine = 1;
	  while ((getline < realSourceFile) == 1) {
            linecount++;
	    if (beginComment($1) && $2=="...") {
              # if (printLine) printf("// ...\n") > tmpFile;
              if (printLine) {
                for (i = 1; i < noblank; ++i) 
		  sourcelines[lines] = sourcelines[lines] " ";
                sourcelines[lines++] = sourcelines[lines] "...";
              }
	      printLine = (printLine + 1)%2;
	      continue;
	    }
            if (beginComment($1) && $2=="BEGIN") {
              skip_mark = $3;
	      noblank = -1; # we want to know the indentation of the following
              while (((getline < realSourceFile) == 1) &&              #  line.
		     !(beginComment($1) && $2=="END" && $3==skip_mark)) {
                linecount++;
                if (noblank == -1 && NF != 0) {
                  tabString = "";
                  for (i = 0; i < TAB; i++) tabString = tabString " ";
                  gsub(/\t/, tabString);
                  noblank = match($0, /[^ \t]/);
                }
              }
	      linecount++;
              if (printLine) {
                for (i = 1; i < noblank; ++i)
		  sourcelines[lines] = sourcelines[lines] " ";
                # The first one is for "pdlsthighlight" in ProgDoc mode.
                if (CPP2L_CMD == "pdlsthighlight") {
                  # In pdlsthighlight, LaTeX commands are escaped by:
                  # &@!& ... LaTeX ... @
                  sourcelines[lines++] = sourcelines[lines]\
                    "&@!&<{\\itshape see Listing \\ref{"\
                    makeLabel(skip_mark) "} on page \\pageref{"\
                    makeLabel(skip_mark) "}}>%\n@";
                }
                # The second one is for "pdlsthighlight" in nuWeb mode.
                # This may be achieved by setting the highlighter to 
                # "pdlsthighlight -n".
                else if (index(CPP2L_CMD, "pdlsthighlight") == 1) {
                  # In pdlsthighlight, LaTeX commands are escaped by:
                  # &@!& ... LaTeX ... @
                  sourcelines[lines++] = sourcelines[lines]\
                    "&@!&{\\normalfont$\\langle$Expanded in listing \\ref{"\
                    makeLabel(skip_mark) "}, page \\pageref{"\
                    makeLabel(skip_mark) "}$\\rangle$}%\n@";
                }
		else { # if (CPP2L_CMD=="pdhighlight") {
                  # In pdhighlight, LaTeX commands are escaped by:
                  # /*? ... LaTeX ... ?*/
                  sourcelines[lines++] = sourcelines[lines]\
		    "/*?\\begin{latexonly}<{\\itshape see Listing \\ref{"\
		    makeLabel(skip_mark) "} on page \\pageref{"\
		    makeLabel(skip_mark) "}}>\\end{latexonly}"\
		    "\\begin{htmlonly}%\n<{\\itshape \\hyperref{\\itshape "\
		    " see Listing }{}{}{" makeLabel(skip_mark) "}\\ref{"\
		    makeLabel(skip_mark) "}}>%\n\\end{htmlonly}?*/";
                }
              }
              continue;  
            }
            if (beginComment($1) && $2=="END" && $3==example) { 
              lineend = linecount - 1;
              break;
            }
            if (NF != 0) { # We compute 'noblank# only for non-empty lines!
	      noblank = match($0, /[^ \t]/);
              if (noblank < blankcount) blankcount = noblank;
            }
	    # if (printLine) printf("%s\n", $0) > tmpFile;
            if (printLine) { 
              sourcelines[lines++] = $0;
              if (length() > maxLineWidth) maxLineWidth = length();
            }
	  }
	  break;
	}
	else if (beginComment($1) && $2=="BEGIN") {
          beginlines[++begincount] = $3;
        }
	else if (beginComment($1) && $2=="END") {
          if (beginlines[begincount] == $3) begincount--;
        }
      }
      if (lines == 1) error(6) 
      else {
        if (wrap) maxLineWidth = wrap;
        else maxLineWidth = maxLineWidth - blankcount + 1;
        for (x = 1; x < lines; x++) {
          tmpStr = substr(sourcelines[x], blankcount, length(sourcelines[x]));
          if (wrap) tmpStr = wrapLine(tmpStr);
          printf("%s\n", tmpStr) > tmpFile;
        }
      }
      # delete sourcelines; # delete <array> is GNU specific
      for (j in sourcelines) delete sourcelines[j];
      close(realSourceFile);
      close(tmpFile);
      FS = oldFS;
      formatCmd = CPP2L_CMD " -t ";
      if (useLongtable) formatCmd = formatCmd " -w ";
      if (linesep) formatCmd = formatCmd " -l " linesep;
      if (fontsize) formatCmd = formatCmd " -s " fontsize;
      if (fontname) formatCmd = formatCmd " -F " fontname;
      if (fontenc) formatCmd = formatCmd " -E " fontenc;
      if (label) formatCmd = formatCmd " -L " label;
      if (TAB != 8) formatCmd = formatCmd " -T " TAB;
      if (begincount) 
	formatCmd = formatCmd " -R " makeLabel(beginlines[begincount]);
      # delete beginlines; # delete <array> is GNU specific
      for (j in beginlines) delete beginlines[j];
      if (linenr && example!="ALL") 
	formatCmd = formatCmd " -b " linebegin " -e " lineend;
      if (listing) {
        if (caption) formatCmd = formatCmd " -A \"" caption "\"";
        formatCmd = formatCmd " -N \"" sourceFile "\"";
      }
      if (listingType) formatCmd = formatCmd " -W " listingType;
      # '-z' takes an optional argument, and there must be NO space between
      # the short paramter '-z' and the optional argument. Another possiblity
      # is to use the long paramter '--center' where '=' acts as a separator
      # between the long paramter and its optional argument.
      if (center) formatCmd = formatCmd " -z" maxLineWidth " ";
      if (underline) formatCmd = formatCmd " -u ";
      if (type) formatCmd = formatCmd " -m " type; #" --" typeName(type);
      formatCmd = formatCmd " -D " makeLabel(example);
      formatCmd = formatCmd " " tmpFile;
      print formatCmd;
      if (CPP2L_CMD=="pdhighlight") {
        while ((formatCmd | getline) == 1) {
          if (center && !useLongtable && 
              (match($0, /\\mbox{}\\\\$/) )) {
            $0 = "\\makebox[\\pdCaptionLength][l]{" substr($0, 1, RSTART)\
	         "mbox{}}\\\\"; 
          }
	  printf("%s\n", $0) > outFile;
	}
      }
      else {
        while ((formatCmd | getline) == 1) printf("%s\n", $0) > outFile;
      }
      close(formatCmd);
    }
    else if (lineBeginsWith($0, "\\sourcebegin")) {
      sourceinput = $0;
      # This is only a simple heuristic. We read at most DELTA lines in order
      # to find all the arguments required for the \sourcebegin command.
      for (i = 0; i < DELTA; ++i) {
        # Protect all braces inside strings enclosed ba apostrophs. For details
        # see the 'parseArguments()' function.
        while (match(sourceinput, /'([^']|\\')*'/)) {
          replA[countA] = gensub(/\\'/, "'", "g", 
                                 substr(sourceinput, RSTART, RLENGTH));#"
           sub(/'([^']|\\')*'/, "_____" countA "_____", sourceinput);
          countA++;
        }

	openb = gsub(/\{/, "{", sourceinput);
	closeb = gsub(/\}/, "}", sourceinput)
	if ((gsub(/\[/, "[", sourceinput) != 1) || 
            (gsub(/\]/, "]", sourceinput) != 1) ||
            (openb != closeb)) {
	  if ((getline < ARGV[1]) != 1) error(5);
	  gsub(/^[ \t]*/," "); # Cut trailing whitespace from input line.
	  sourceinput = sourceinput $0;
	}  
      }
      gsub(/^.*$/,sourceinput); # Replace $0 by 'sourceinput'.
      print $0;
      if (gsub(/\[/, "[") != gsub(/\]/, "]")) error(5);
      if (gsub(/\[/, "[") == 1) {
        # Here we restore the parts of $2 which may have been replaced before.
        # This is necessary because 'parseArguments()' uses the global variable
        # $2 directly.
        for (i = 0; i < countA; i++) {
          ind = index($2, "_____" i "_____");
          if (ind) $2 = substr($2, 0, ind-1) replA[i] substr($2, ind + 11);
        }
	parseArguments();
      }
      if (gsub(/{/, "{", sourceinput) >= 1) {  # We may have more braces 
	firstb = index(sourceinput, "{") + 1;  #                   in the text!
	sourceFile = substr(sourceinput, firstb, length(sourceinput) - firstb);
      }
      else {
        if (caption) sourceFile = caption;
	else sourceFile = ""; # We must reset 'sourceFile' since it may be 
      }                       # preset from the last 'sourceinput' command.
      maxLineWidth = 0;
      while ((getline < ARGV[1]) == 1) {
        if (lineBeginsWith($0, "\\sourceend")) break;
        if (length() > maxLineWidth) maxLineWidth = length();
	printf("%s\n", $0) > tmpFile;
      }
      close(tmpFile);
      formatCmd = CPP2L_CMD " -t ";
      if (useLongtable) formatCmd = formatCmd " -w ";
      if (linesep) formatCmd = formatCmd " -l " linesep;
      if (fontsize) formatCmd = formatCmd " -s " fontsize;
      if (fontname) formatCmd = formatCmd " -F " fontname;
      if (fontenc) formatCmd = formatCmd " -E " fontenc;
      if (label) formatCmd = formatCmd " -L " label;
      if (linenr) formatCmd = formatCmd " -b " linebegin " -e " lineend;
      # Notice that we need the -N option in order to just create the caption.
      # The difference between -A and -N is that -A has precedence over -N and
      # that the argument of -A will be not quoted like the on of -N which is
      # assumend to be a file name.
      if (listing) {
        formatCmd = formatCmd " -N \"" sourceFile "\"";
        formatCmd = formatCmd " -A \"" sourceFile "\"";
      }
      if (listingType) formatCmd = formatCmd " -W " listingType;
      # '-z' takes an optional argument, and there must be NO space between
      # the short paramter '-z' and the optional argument. Another possiblity
      # is to use the long paramter '--center' where '=' acts as a separator
      # between the long paramter and its optional argument.
      if (center) formatCmd = formatCmd " -z" maxLineWidth " ";
      if (underline) formatCmd = formatCmd " -u ";
      if (type) formatCmd = formatCmd " -m " type; #" --" typeName(type);
      formatCmd = formatCmd " " tmpFile;
      print formatCmd;
      if (CPP2L_CMD=="pdhighlight") {
        while ((formatCmd | getline) == 1) {
          if (center && !useLongtable && match($0, /\\mbox{}\\\\$/)) {
            $0 = "\\makebox[\\pdCaptionLength][l]{" substr($0, 1, RSTART)\
	         "mbox{}}\\\\"; 
          }
	  printf("%s\n", $0) > outFile;
	}
      }
      else {
        while ((formatCmd | getline) == 1) printf("%s\n", $0) > outFile;
      }
      close(formatCmd);
    }
    else printf("%s\n", $0) > outFile;
  }
}
